BOTH_SF_CONNECTION_AND_SNOWPARK_SESSION_SPECIFIED = (
    "sf_connection and snowpark_session cannot be specified at the same time."
)
NO_SF_CONNECTION_OR_SNOWPARK_SESSION = "sf_connection or snowpark_session must be provided"
NO_SF_CONNECTION_OR_SNOWPARK_DATAFRAME = "Either snowpark_dataframe or sf_connection should be non-empty."

# FileSetError
FILESET_ALREADY_EXISTS = "FileSet with name {} has already existed."
