print("import: 'snowflake.ml'")
import snowflake.ml

print("import: 'snowflake.ml.fileset'")
import snowflake.ml.fileset

print("import: 'snowflake.ml.model'")
import snowflake.ml.model

print("import: 'snowflake.ml.modeling'")
import snowflake.ml.modeling

print("import: 'snowflake.ml.registry'")
import snowflake.ml.registry

print("import: 'snowflake.ml.training_dataset'")
import snowflake.ml.training_dataset

print("import: 'snowflake.ml.utils'")
import snowflake.ml.utils

